/*
 * Copyright (c) 2012 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "velodyneTarget.h"

/* System headers */
#include <float.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h> /* Necessary for velodyneLib.h */
#include <sys/types.h> /* Necessary for velodyneLib.h */

/* External headers */
#include <velodyne/velodyneLib.h>

/* Genom headers */
#include <portLib.h>

/* Directives */
#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif

#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

#define	MAX_SEED_NUMBER 100
#define SEEDSRATIO 0.1

/* Functions */
static ACTIVITY_EVENT velodyne_targetRansacDetection(
    velodyneTargetDetectionParams*, velodyneTargetParams*, velodyneTarget*);
static void velodyne_decideTargetStatus(velodyneTargetParams*, velodyneTarget*);
static ACTIVITY_EVENT velodyne_updateTargetCharacteristics(velodyneTarget*);
static ACTIVITY_EVENT velodyne_filterTargetPoints(velodyneTarget*);
static void velodyne_printTargetCharacteristics(velodyneTarget*);
static int velodyne_compareClusters(const void*, const void*);

/***********************************************************************
 *
 * Process to detect and estimate the target position 
 *
 ***********************************************************************/
ACTIVITY_EVENT
velodyne_analyzeTarget (velodyneTargetParams*		targetParams,
			velodyneTargetDetectionParams*	detectionParams,
			velodyneTarget*			target)
{
  return (velodyne_targetRansacDetection (detectionParams, targetParams, target));
  /* return END; */
}

/***********************************************************************
 *
 * On the basis of all the 3D points of the target, estimate its
 * center of gravity and bounds
 *
 ***********************************************************************/
ACTIVITY_EVENT
velodyne_targetRansacDetection (velodyneTargetDetectionParams*	detectionParams,
				velodyneTargetParams*		targetParams,
				velodyneTarget*			target)
{
  int		seedNumber; 	/* The number of seeds */
  int		randomNumber;
  velodyne3DPoint*	point;
  velodyne3DPoint*	seedPoint;
  velodyne3DPoint*	targetPoint;
  float			xSeed, ySeed, zSeed;
  int			index;
  velodyneTarget*	seeds;
  velodyneTarget*	seedTarget;
  velodyneTarget	foundTarget;
  int			i, j;
  int			maxSeedSize;
  int			indexMaxSeedSize;


  srand (time(NULL));

  seeds = (velodyneTarget*) malloc (MAX_SEED_NUMBER * sizeof(velodyneTarget));
  if (seeds == NULL) {
    fprintf (stderr, "velodyne: can't allocate targets structure\n");
    return FAIL;
  }
  memset ( (velodyneTarget*) seeds, 0, MAX_SEED_NUMBER * sizeof(velodyneTarget));
  memset (&foundTarget, 0, sizeof(velodyneTarget));

  /* Computes the number of seeds */
  seedNumber = (int) (SEEDSRATIO * ((float) target->detectedTargetPoints));
  seedNumber = MIN (MAX_SEED_NUMBER, seedNumber);
  if (detectionParams->verbosity == VELODYNE_VERY_VERBOSE) {
    printf ("velodyne: %d detected reflective points\n", 
	    target->detectedTargetPoints);
    printf ("velodyne: %d seeds used for the cluster detection process\n", 
	    seedNumber);
  }
  
  if (seedNumber == 0) {
    target->status = VELODYNE_TARGET_NOT_DETECTED;
    if (detectionParams->verbosity != VELODYNE_QUIET) 
      printf ("velodyne: not enough reflective points detected\n");
    free (seeds);
    return END;
  }

  /* Fill the seeds */
  for (i = 0; i < seedNumber; i++) {
    /* Select a seed point, making sure it has not already been
       selected */
    randomNumber = rand();
    index = target->detectedTargetPoints * (((float) randomNumber) / RAND_MAX);
    if (index == target->detectedTargetPoints)
      index--; /* To account for the case where rand() returns RAND_MAX */ 

    while ((target->targetPoints + index)->status == VELODYNE_BAD_POINT) {
      /* skip an already selected seed */
      randomNumber = rand();
      index = target->detectedTargetPoints * (((float) randomNumber) / RAND_MAX);
    }

    seedPoint = target->targetPoints + index;
    seedPoint->status = VELODYNE_BAD_POINT; /* To avoid reselecting
					       it as a seed next time */
    xSeed = seedPoint->coordinates[0];
    ySeed = seedPoint->coordinates[1];
    zSeed = seedPoint->coordinates[2];

    /* And fill the associated target structure */
    seedTarget = seeds + i;
    seedTarget->detectedTargetPoints = 0;
    for (j = 0; j < target->detectedTargetPoints; j++) {
      point = target->targetPoints + j;
      if ( (fabs(xSeed - point->coordinates[0]) < targetParams->radius)
	  && (fabs(ySeed - point->coordinates[1]) < targetParams->radius)
	  && (fabs(zSeed - point->coordinates[2]) < targetParams->radius)) {
	/* This point is in the box, copy it for further analysis */
	/* memcpy ((seedTarget->targetPoints + seedTarget->detectedTargetPoints),  */
	/* 	point, sizeof(velodyne3DPoint)); */
	targetPoint = seedTarget->targetPoints + seedTarget->detectedTargetPoints;
	targetPoint->coordinates[0] = point->coordinates[0];
	targetPoint->coordinates[1] = point->coordinates[1];
	targetPoint->coordinates[2] = point->coordinates[2];
	targetPoint->intensity = point->intensity;
	targetPoint->status = VELODYNE_GOOD_3DPOINT;
	seedTarget->detectedTargetPoints += 1;
      }
    }
  }

  /* Now find the seed that contains the maximum nb of points, update
     its characteristics and prints them */
  indexMaxSeedSize = -1;
  maxSeedSize = 0;
  for (i = 0; i < seedNumber; i++) {
    if ((((seeds + i)->detectedTargetPoints) > maxSeedSize)
	&& ((seeds + i)->status != VELODYNE_TARGET_DETECTED)) {
      maxSeedSize = (seeds + i)->detectedTargetPoints;
      indexMaxSeedSize = i;
      memcpy (&foundTarget, seeds + i, sizeof (velodyneTarget));
    }
  }

  velodyne_updateTargetCharacteristics (&foundTarget);
  /* Decide whether the target is detected or not */
  velodyne_decideTargetStatus (targetParams, &foundTarget);
  switch (foundTarget.status) {
  case VELODYNE_TARGET_DETECTED:
    if (detectionParams->verbosity != VELODYNE_QUIET) {
      printf ("velodyne: target DETECTED\n");
      velodyne_printTargetCharacteristics (&foundTarget);
    }
    break;
  case VELODYNE_TARGET_NOT_ENOUGH_POINTS:
    if (detectionParams->verbosity != VELODYNE_QUIET) {
      printf ("velodyne: not enough points in the largest cluster\n");
      velodyne_printTargetCharacteristics (&foundTarget);
    }
    break;
  case VELODYNE_TARGET_NOT_WIDE_ENOUGH:
    if (detectionParams->verbosity != VELODYNE_QUIET) {
      printf ("velodyne: the largest cluster is not wide enough\n");
      velodyne_printTargetCharacteristics (&foundTarget);
    }
    break;
  case VELODYNE_TARGET_TOO_WIDE:
    if (detectionParams->verbosity != VELODYNE_QUIET) {
      printf ("velodyne: the largest cluster is too wide\n");
      velodyne_printTargetCharacteristics (&foundTarget);
    }
    break;
  case VELODYNE_TARGET_NOT_HIGH_ENOUGH:
    if (detectionParams->verbosity != VELODYNE_QUIET) {
      printf ("velodyne: the largest cluster is not high enough\n");
      velodyne_printTargetCharacteristics (&foundTarget);
    }
    break;
  case VELODYNE_TARGET_TOO_HIGH:
    if (detectionParams->verbosity != VELODYNE_QUIET) {
      printf ("velodyne: the largest cluster is too high\n");
      velodyne_printTargetCharacteristics (&foundTarget);
    }
    break;
  case VELODYNE_TARGET_NOT_DETECTED:
    if (detectionParams->verbosity != VELODYNE_QUIET) 
      printf ("velodyne: no cluster of reflective points detected\n");
    break;
  default:
    break;
  }

  /* Print the characteristics of all the clusters for debugging
     purposes */
  if (detectionParams->verbosity == VELODYNE_VERY_VERBOSE) {
    /* Sort them by number of target points */
    qsort (seeds, seedNumber, sizeof(velodyneTarget), velodyne_compareClusters);
    printf ("\n\nvelodyne: printing info on all the detected target clusters\n");
    for (i = 0; i < seedNumber; i++) {
      printf ("velodyne: cluster %d\n", i);
      velodyne_updateTargetCharacteristics (seeds + i);
      velodyne_decideTargetStatus (targetParams, seeds + i);
      velodyne_printTargetCharacteristics (seeds + i);
    }
  }

  /* Copy the detected target */
  memcpy (target, &foundTarget, sizeof(velodyneTarget));

  free(seeds);

  return END;
}

/***********************************************************************
 *
 * Decide whether the target is detected or not 
 *
 ***********************************************************************/
void 
velodyne_decideTargetStatus (velodyneTargetParams*		targetParams,
			     velodyneTarget*			target)
{
  double	distance;	/* Distance between the target COG and
				   the sensor */
  double	nbPointsMin;	/* Minimun number of points required
				   to declare the target detected */

  distance = sqrt ((target->cog.coordinates[0] * target->cog.coordinates[0])
		   + (target->cog.coordinates[1] * target->cog.coordinates[1])
		   + (target->cog.coordinates[2] * target->cog.coordinates[2]));
  
  /* Compute the minimum number of requested points with the distance.
  The target surface is (very roughly) maxWidth * maxHeight. The
  horizontal angular resolution of the velodyne is roughly 0.1 degree
  (at a 5Hz rotation rate) and the vertical angular resolution is
  roughly 0.15 degrees. Hence 1m2 at 1m distance generates roughly
  2400 points. The points density therefore decreases according to
  2600/(distance^2). Roughly. */
  nbPointsMin = (targetParams->minWidth * targetParams->minHeight) 
    * targetParams->nbRequestedPoints / (distance * distance);

  /* Consider the target detected, the check criteria one by one */
  target->status = VELODYNE_TARGET_DETECTED;
  if (target->detectedTargetPoints < nbPointsMin)
    target->status = VELODYNE_TARGET_NOT_ENOUGH_POINTS;
  if (target->radius < targetParams->minWidth)
    target->status = VELODYNE_TARGET_NOT_WIDE_ENOUGH;
  if (target->radius > targetParams->maxWidth)
    target->status = VELODYNE_TARGET_TOO_WIDE;
  if ((target->max.coordinates[2] - target->min.coordinates[2]) 
      < targetParams->minHeight)
    target->status = VELODYNE_TARGET_NOT_HIGH_ENOUGH;
  if ((target->max.coordinates[2] - target->min.coordinates[2]) 
      > targetParams->maxHeight)
    target->status = VELODYNE_TARGET_TOO_HIGH;
  
  return;
}


/***********************************************************************
 *
 * On the basis of all the 3D points of the target, estimate its
 * center of gravity and bounds
 *
 ***********************************************************************/
ACTIVITY_EVENT

velodyne_updateTargetCharacteristics (velodyneTarget*		target)
{
  int			i;
  float			x2, y2, z2;
  velodyne3DPoint*	point;
  int			nbPoints = 0;

  /* Brute force approach to compute the mean and standard deviations
     -- there are much more elegant ways to compute this, but we will
     see later */
  target->cog.coordinates[0] = 0;
  target->cog.coordinates[1] = 0;
  target->cog.coordinates[2] = 0;
  target->stdDev.coordinates[0] = 0;
  target->stdDev.coordinates[1] = 0;
  target->stdDev.coordinates[2] = 0;
  target->max.coordinates[0] = -FLT_MAX;
  target->max.coordinates[1] = -FLT_MAX;
  target->max.coordinates[2] = -FLT_MAX;
  target->min.coordinates[0] = FLT_MAX;
  target->min.coordinates[1] = FLT_MAX;
  target->min.coordinates[2] = FLT_MAX;

  nbPoints = 0;
  target->meanReflectance = 0.0;
  for (i = 0; i < target->detectedTargetPoints; i++) {
    point = target->targetPoints + i;
    if (point->status == VELODYNE_GOOD_3DPOINT) {
      nbPoints += 1;

      target->meanReflectance += point->intensity;

      target->cog.coordinates[0] += point->coordinates[0];
      target->cog.coordinates[1] += point->coordinates[1];
      target->cog.coordinates[2] += point->coordinates[2];
      
      target->max.coordinates[0] 
	= MAX(point->coordinates[0], target->max.coordinates[0]);
      target->max.coordinates[1] 
	= MAX(point->coordinates[1], target->max.coordinates[1]);
      target->max.coordinates[2] 
	= MAX(point->coordinates[2], target->max.coordinates[2]);
      
      target->min.coordinates[0] 
	= MIN(point->coordinates[0], target->min.coordinates[0]);
      target->min.coordinates[1] 
	= MIN(point->coordinates[1], target->min.coordinates[1]);
      target->min.coordinates[2] 
	= MIN(point->coordinates[2], target->min.coordinates[2]);
    }
  }
  target->meanReflectance /= nbPoints;
  target->cog.coordinates[0] /= nbPoints;
  target->cog.coordinates[1] /= nbPoints;
  target->cog.coordinates[2] /= nbPoints;
  target->radius = sqrt ( (target->max.coordinates[0] - target->min.coordinates[0])*(target->max.coordinates[0] - target->min.coordinates[0]) + (target->max.coordinates[1] - target->min.coordinates[1])*(target->max.coordinates[1] - target->min.coordinates[1]));

  x2 = 0;
  y2 = 0;
  z2 = 0;
  nbPoints = 0;
  for (i = 0; i < target->detectedTargetPoints; i++) {
    point = target->targetPoints + i;
    if (point->status == VELODYNE_GOOD_3DPOINT) {
      nbPoints += 1;
      x2 += (point->coordinates[0] * point->coordinates[0]);
      y2 += (point->coordinates[1] * point->coordinates[1]);
      z2 += (point->coordinates[2] * point->coordinates[2]);
    }
  }
  target->stdDev.coordinates[0] 
    = sqrt (x2/nbPoints - 
	    (target->cog.coordinates[0]*target->cog.coordinates[0]));
  target->stdDev.coordinates[1] 
    = sqrt (y2/nbPoints - 
	    (target->cog.coordinates[1]*target->cog.coordinates[1]));
  target->stdDev.coordinates[2] 
    = sqrt (z2/nbPoints - 
	    (target->cog.coordinates[2]*target->cog.coordinates[2]));

  return END;
}


/***********************************************************************
 *
 * On the basis of all the computed target center of gravity (cog) and
 * standard deviation, remove all the points whose at least one
 * coordinate is further than 3*stdDev of the cog points
 *
 ***********************************************************************/
ACTIVITY_EVENT

velodyne_filterTargetPoints (velodyneTarget*		target)
{
  int			i;
  velodyne3DPoint*	point;

  for (i = 0; i < target->detectedTargetPoints; i++) {
    point = target->targetPoints + i;
    if (point->status == VELODYNE_GOOD_3DPOINT) {
      if ( (fabs(point->coordinates[0] - target->cog.coordinates[0]) 
	    > 3.0 * target->stdDev.coordinates[0]) ||
	   (fabs(point->coordinates[1] - target->cog.coordinates[1]) 
	    > 3.0 * target->stdDev.coordinates[1]) ||
	   (fabs(point->coordinates[2] - target->cog.coordinates[2]) 
	    > 3.0 * target->stdDev.coordinates[2]) ) {
	point->status = VELODYNE_BAD_3DPOINT;
	target->detectedTargetPoints--;
      }
    }
  }
  return END;
}

/***********************************************************************
 *
 * Just prints the characteristics of the target
 *
 ***********************************************************************/

void
velodyne_printTargetCharacteristics ( velodyneTarget*			target)
{
  printf ("\t\tcentered on (%.2f, %.2f, %.2f), %d points, meanReflectance = %.2f\n"
	  "\t\tRadius : %.2f meters\n"
	  "\t\tX bounds : %.2f < x < %.2f, delta = %.2f, stDev = %.2f\n"
	  "\t\tY bounds : %.2f < y < %.2f, delta = %.2f, stDev = %.2f\n"
	  "\t\tZ bounds : %.2f < z < %.2f, delta = %.2f, stDev = %.2f\n",
	  target->cog.coordinates[0],
	  target->cog.coordinates[1],
	  target->cog.coordinates[2],
	  target->detectedTargetPoints,
	  target->meanReflectance,
	  target->radius,
	  target->min.coordinates[0], target->max.coordinates[0], 
	  target->max.coordinates[0] - target->min.coordinates[0], 
	  target->stdDev.coordinates[0],
	  target->min.coordinates[1], target->max.coordinates[1],
	  target->max.coordinates[1] - target->min.coordinates[1], 
	  target->stdDev.coordinates[1],
	  target->min.coordinates[2], target->max.coordinates[2], 
	  target->max.coordinates[2] - target->min.coordinates[2], 
	  target->stdDev.coordinates[2]);
  return;
}

/***********************************************************************
 *
 * Remove from the 3D image all the points that fall within the X and
 * Y bounds of the target  
 *
 ***********************************************************************/

void
velodyne_eraseTargetPoints (velodyne3DImage*	im3d, 
			    velodyneTarget*	target)
{
  int			i, j;
  velodyne3DPoint*	point;
  int nbErasedPoints = 0;

  for (i = 0; i < VELODYNE_3D_IMAGE_HEIGHT; i++) {
    for (j = 0; j < VELODYNE_3D_IMAGE_WIDTH; j++) {
      point = velodyne3DImageAt(im3d, i, j);
      if (point->status == VELODYNE_GOOD_3DPOINT) {
	if ((point->coordinates[0] >= target->min.coordinates[0])
	    && (point->coordinates[0] <= target->max.coordinates[0])
	    && (point->coordinates[1] >= target->min.coordinates[1])
	    && (point->coordinates[1] <= target->max.coordinates[1])) {
	  point->status = VELODYNE_BAD_3DPOINT;
	  nbErasedPoints++;
	}
      }
    }
  }
  return;
}

/* Used to sort the clusters before printing all of them */
static int
velodyne_compareClusters (const void* cluster1,
			  const void* cluster2)
{
  if ( ((velodyneTarget*) cluster1)->detectedTargetPoints >
       ((velodyneTarget*) cluster2)->detectedTargetPoints )
    return -1;
  else
    return 1;
}

/* Target tracking */
void
velodyneTrackTarget (
    velodyneTarget* target,
    velodyneTargetTracking* targetTracking,
    velodyneTargetTrackingParams* targetTrackingParams,
    velodyneTargetTrackingFileOutputParams* targetTrackingFileOutputParams,
    float targetCOGX, float targetCOGY,
    int* report)
{
  velodyneTarget* trackedTarget;
  float targetMovement;
  int i;
  FILE* outputFile;
  if (targetTracking->numTargets >= VELODYNE_DEFAULT_MAX_NUM_TARGETS)
    return;
  for (i = 0; i < targetTracking->numTargets; i++) {
    targetMovement
        = (targetCOGX - (targetTracking->targets[i]).cog.coordinates[0])
          * (targetCOGX - (targetTracking->targets[i]).cog.coordinates[0])
          + (targetCOGY - (targetTracking->targets[i]).cog.coordinates[1])
          * (targetCOGY - (targetTracking->targets[i]).cog.coordinates[1]);
    if (targetMovement <= targetTrackingParams->maxTargetMovement)
      return;
  }
  trackedTarget = &(targetTracking->targets[targetTracking->numTargets]);
  memcpy (trackedTarget, target, sizeof (velodyneTarget));
  trackedTarget->cog.coordinates[0] = targetCOGX;
  trackedTarget->cog.coordinates[1] = targetCOGY;
  trackedTarget->cog.coordinates[2] = 0.0;
  targetTracking->numTargets += 1;
  if (targetTrackingFileOutputParams->enableFileOutput == VELODYNE_TRUE) {
    outputFile = fopen(targetTrackingFileOutputParams->outputFileName, "a");
    if (outputFile == NULL) {
      *report = S_velodyne_CANNOT_OPEN_FILE;
      return;
    }
    fprintf(outputFile, "%.2f %.2f\n", targetCOGX, targetCOGY);
    fclose(outputFile);
  }
  return;
}
