/*
 * Copyright (c) 2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 ** velodyneExecTaskCodels.c
 **
 ** Codels used by the execution task ExecTask
 **
 **/

/* System headers */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* External headers */
#include <libimages3d.h>
#include <t3d/t3d.h>

/* Genom headers */
#include <portLib.h>

/* External Genom headers */
#include <libeuler/pomEuler.h>

/* Internal headers */
#include "server/velodyneHeader.h"
#include "velodyne3DImage.h"
#include "velodyneTarget.h"
#include "velodyneBuffer.h"

/* Public variables */
VELODYNE_INTERFACE* interface;
velodyneReadTask task;
POSTER_ID sensorPosPosterId;

/* Private variables */
static FILE* posFileId;
static velodyne3DImageParams imageParams;
static velodyneTarget target;
static velodyneTargetTracking* targetTracking;
static velodyneMode mode;
static char readImageFileName[VELODYNE_MAX_STRLEN];
static char readPosFileName[VELODYNE_MAX_STRLEN];

/* Posters */
static velodyne3DImage* threeDImage;
static GENPOS_TRAJ_POINTS* targetAbsolutePosition;
static GENPOS_TRAJ_POINTS* targetRelativePosition;
static GENPOS_TRAJ_POINTS* targetTrackingPoster;

/*------------------------------------------------------------------------------
 * ExecTask - Start
 *
 * Description: sets IDS members to their default values and initializes
 * external variables
 *
 * Returns: OK or ERROR
 */
STATUS
velodyneExecTaskStart(int* report)
{
  /* sensorToMain */
  SDI_F->sensorToMain.yaw = VELODYNE_DEFAULT_SENSOR_TO_MAIN_YAW;
  SDI_F->sensorToMain.pitch = VELODYNE_DEFAULT_SENSOR_TO_MAIN_PITCH;
  SDI_F->sensorToMain.roll = VELODYNE_DEFAULT_SENSOR_TO_MAIN_ROLL;
  SDI_F->sensorToMain.x = VELODYNE_DEFAULT_SENSOR_TO_MAIN_X;
  SDI_F->sensorToMain.y = VELODYNE_DEFAULT_SENSOR_TO_MAIN_Y;
  SDI_F->sensorToMain.z = VELODYNE_DEFAULT_SENSOR_TO_MAIN_Z;

  /* targetDetectionParams */
  SDI_F->targetDetectionParams.detectionOn = VELODYNE_DEFAULT_DETECTION_MODE;
  SDI_F->targetDetectionParams.trackingOn = VELODYNE_DEFAULT_TRACKING_MODE;
  SDI_F->targetDetectionParams.verbosity = VELODYNE_DEFAULT_VERBOSITY;
  SDI_F->targetDetectionParams.eraseTargetPoints
      = VELODYNE_DEFAULT_ERASE_TARGET_POINTS;

  /* targetParams */
  SDI_F->targetParams.maxDistance = VELODYNE_DEFAULT_MAX_TARGET_DISTANCE;
  SDI_F->targetParams.zMin = VELODYNE_DEFAULT_TARGET_Z_MIN;
  SDI_F->targetParams.minReflectance = VELODYNE_DEFAULT_MIN_REFLECTANCE;
  SDI_F->targetParams.maxReflectance = VELODYNE_DEFAULT_MAX_REFLECTANCE;
  SDI_F->targetParams.radius = VELODYNE_DEFAULT_TARGET_RADIUS;
  SDI_F->targetParams.nbRequestedPoints = VELODYNE_DEFAULT_NB_REQUESTED_POINTS;
  SDI_F->targetParams.minWidth = VELODYNE_DEFAULT_TARGET_MIN_WIDTH;
  SDI_F->targetParams.maxWidth = VELODYNE_DEFAULT_TARGET_MAX_WIDTH;
  SDI_F->targetParams.minHeight = VELODYNE_DEFAULT_TARGET_MIN_HEIGHT;
  SDI_F->targetParams.maxHeight = VELODYNE_DEFAULT_TARGET_MAX_HEIGHT;

  /* targetTrackingParams */
  SDI_F->targetTrackingParams.maxTargetMovement
      = VELODYNE_DEFAULT_MAX_TARGET_MOVEMENT;
  SDI_F->targetTrackingFileOutputParams.enableFileOutput
      = VELODYNE_DEFAULT_ENABLE_TARGET_TRACKING_FILE_OUTPUT;
  strncpy (SDI_F->targetTrackingFileOutputParams.outputFileName,
           VELODYNE_DEFAULT_TARGET_TRACKING_OUTPUT_FILE_NAME,
           VELODYNE_MAX_STRLEN);

  /* saveParams */
  strncpy (SDI_F->saveParams.baseName, VELODYNE_DEFAULT_BASE_NAME,
           VELODYNE_MAX_STRLEN);
  SDI_F->saveParams.num = VELODYNE_DEFAULT_NUM;

  /* saveIm3DParams */
  strncpy (SDI_F->saveIm3DParams.baseName, VELODYNE_DEFAULT_BASE_NAME_IM3D,
           VELODYNE_MAX_STRLEN);
  SDI_F->saveIm3DParams.num = VELODYNE_DEFAULT_NUM;

  /* Others */
  SDI_F->running = 0;

  /* Public variables */
  interface = NULL;

  /* Private variables */
  posFileId = NULL;
  targetTracking = malloc (sizeof (velodyneTargetTracking));
  if (targetTracking == NULL) {
    fprintf (stderr, "velodyne: cannot allocate target tracking structure\n");
    return ERROR;
  }
  memset (targetTracking, 0, sizeof (velodyneTargetTracking));
  strcpy(readImageFileName, "");
  strcpy(readPosFileName, "");

  /* Posters */
  threeDImage = malloc (sizeof (velodyne3DImage));
  if (threeDImage == NULL) {
    fprintf (stderr, "velodyne: cannot allocate 3D image structure\n");
    return ERROR;
  }
  targetAbsolutePosition = malloc (sizeof (GENPOS_TRAJ_POINTS));
  if (targetAbsolutePosition == NULL) {
    fprintf (stderr,
             "velodyne: cannot allocate target absolute position structure\n");
    return ERROR;
  }
  memset (targetAbsolutePosition, 0, sizeof (GENPOS_TRAJ_POINTS));
  targetRelativePosition = malloc (sizeof (GENPOS_TRAJ_POINTS));
  if (targetRelativePosition == NULL) {
    fprintf (stderr,
             "velodyne: cannot allocate target relative position structure\n");
    return ERROR;
  }
  memset (targetRelativePosition, 0, sizeof (GENPOS_TRAJ_POINTS));
  targetTrackingPoster = malloc (sizeof (GENPOS_TRAJ_POINTS));
  if (targetTrackingPoster == NULL) {
    fprintf (stderr,
             "velodyne: cannot allocate target tracking poster structure\n");
    return ERROR;
  }
  memset (targetTrackingPoster, 0, sizeof (GENPOS_TRAJ_POINTS));

  return OK;
}

/*------------------------------------------------------------------------------
 * ExecTask - End
 *
 * Description: handles the termination of resources used by external variables
 *
 * Returns: OK
 */

STATUS
velodyneExecTaskEnd(void)
{
  /* Public variables */
  if (interface != NULL)
    velodyne_close (interface);
  if (posFileId != NULL)
    fclose (posFileId);
  /* Private variables */
  free (targetTracking);
  /* Posters */
  free (threeDImage);
  free (targetAbsolutePosition);
  free (targetRelativePosition);
  free (targetTrackingPoster);
  return OK;
}

/*------------------------------------------------------------------------------
 * Init - Main
 *
 * Description: opens the interface and connects to the position input.
 *
 * Reports: S_velodyne_BAD_INIT_PARAMS
 *          S_velodyne_CANNOT_OPEN_INTERFACE
 *          S_velodyne_CANNOT_UPDATE_3D_IMAGE_POSTER
 *          S_velodyne_CANNOT_FIND_POS_POSTER
 *          S_velodyne_CANNOT_READ_POS_POSTER
 *          S_velodyne_CANNOT_OPEN_POS_FILE
 *          S_velodyne_BAD_POS_FILE_FORMAT
 *
 * Returns: ETHER
 */
ACTIVITY_EVENT
velodyneInitMain(velodyneInitParams* initParams, int* report)
{
  VELODYNE_MODEL model;
  POM_SENSOR_POS pos;
  velodyneReaderResult posFileOpenResult;

  /* Sets the 3D image's members and updates the poster. */
  switch (initParams->model) {
  case VELODYNE_HDL32:
    threeDImage->height = VELODYNE_NUM_LASERS_32;
    threeDImage->maxScanWidth = VELODYNE_MAX_NUM_PACKETS_PER_REVOLUTION_32
                                * VELODYNE_NUM_SHOTS_OF_A_LASER_PER_PACKET_32;
    break;
  case VELODYNE_HDL64:
    threeDImage->height = VELODYNE_NUM_LASERS_64;
    threeDImage->maxScanWidth = VELODYNE_MAX_NUM_PACKETS_PER_REVOLUTION_64
                                * VELODYNE_NUM_SHOTS_OF_A_LASER_PER_PACKET_64;
    break;
  default:
    fprintf(stderr, "velodyne: bad model\n");
    *report = S_velodyne_BAD_INIT_PARAMS;
    return ETHER;
  }
  threeDImage->width = 0;
  if (velodyneThreeDImagevelodyne3DImagePosterWrite(
          VELODYNE_THREEDIMAGE_POSTER_ID, threeDImage)
      == ERROR) {
    *report = S_velodyne_CANNOT_UPDATE_3D_IMAGE_POSTER;
    return ETHER;
  }

  /* If in read mode, then there's nothing left to do. */
  if (initParams->mode == VELODYNE_READ)
    return ETHER;

  /* Retrieves the model. */
  switch (initParams->model) {
  case VELODYNE_HDL32:
    model = VELODYNE_32;
    break;
  case VELODYNE_HDL64:
    model = VELODYNE_64;
    break;
  }

  /* Opens the interface. */
  if (interface != NULL) {
    velodyne_close(interface);
    interface = NULL;
  }
  switch (initParams->mode) {
  case VELODYNE_LIVE:
    interface = velodyne_open_network(model);
    break;
  case VELODYNE_FROM_FILE:
    interface = velodyne_open_file(model, initParams->inputFileName);
    break;
  default:
    fprintf(stderr, "velodyne: bad mode\n");
    *report = S_velodyne_BAD_INIT_PARAMS;
    return ETHER;
  }
  if (interface == NULL) {
    fprintf(stderr, "velodyne: cannot open interface\n");
    *report = S_velodyne_CANNOT_OPEN_INTERFACE;
    return ETHER;
  }

  /* Sets the 3D image parameters. */
  switch (model) {
  case VELODYNE_32:
    imageParams.model = VELODYNE_32;
    imageParams.numLinesPerPacket = VELODYNE_NUM_LASERS_32;
    imageParams.numColumnsPerPacket
        = VELODYNE_NUM_SHOTS_OF_A_LASER_PER_PACKET_32;
    break;
  case VELODYNE_64:
    imageParams.model = VELODYNE_64;
    imageParams.numLinesPerPacket = VELODYNE_NUM_LASERS_64;
    imageParams.numColumnsPerPacket
        = VELODYNE_NUM_SHOTS_OF_A_LASER_PER_PACKET_64;
    break;
  }
  if (initParams->enableOrganization == VELODYNE_TRUE) {
    imageParams.organizationEnabled = VELODYNE_TRUE;
    velodyne_get_organization_values(model, imageParams.lineOrder,
                                     imageParams.columnOffsets);
  } else {
    imageParams.organizationEnabled = VELODYNE_FALSE;
  }

  /* Connects to the position input. */
  if (initParams->usePom == VELODYNE_TRUE) {
    switch (initParams->mode) {
    case VELODYNE_LIVE:
      /* Finds the poster. */
      if (velodynePOM_SENSOR_POSPosterFind(initParams->sensorPosPoster,
                                           &sensorPosPosterId)
          == ERROR) {
        fprintf(stderr, "velodyne: cannot find position poster\n");
        *report = S_velodyne_CANNOT_FIND_POS_POSTER;
        return ETHER;
      }
      /* Reads the poster. */
      if (velodynePOM_SENSOR_POSPosterRead(sensorPosPosterId, &pos) == ERROR) {
        fprintf(stderr, "velodyne: cannot read position poster\n");
        *report = S_velodyne_CANNOT_READ_POS_POSTER;
        return ETHER;
      }
      /* Sets the sensor-to-main transformation. */
      SDI_F->sensorToMain = pos.sensorToMain.euler;
      break;
    case VELODYNE_FROM_FILE:
      /* Opens the file. It's not necessary to read it after to set the
         sensor-to-main transformation because in playback mode this is done by
         the request SetSensorToMainPlayback. */
      posFileOpenResult = velodynePosFileOpen(initParams->sensorPosPoster,
                                              &posFileId);
      if (posFileOpenResult != VELODYNE_READER_OK) {
        switch (posFileOpenResult) {
        case VELODYNE_READER_ERR_CANNOT_OPEN_POS_FILE:
          fprintf(stderr, "velodyne: cannot open position file\n");
          *report = S_velodyne_CANNOT_OPEN_POS_FILE;
          break;
        case VELODYNE_READER_ERR_BAD_POS_FILE_FORMAT:
          fprintf(stderr, "velodyne: bad position file format\n");
          *report = S_velodyne_BAD_POS_FILE_FORMAT;
          break;
        default:
          break;
        }
        return ETHER;
      }
      break;
    }
  }

  return ETHER;
}

/*------------------------------------------------------------------------------
 * StartAcquisition - Main
 *
 * Description: starts the data acquisition.
 *
 * Reports: S_velodyne_ACQUISITION_ALREADY_STARTED
 *          S_velodyne_INTERFACE_NOT_OPEN
 *          S_velodyne_NOT_ENOUGH_RESSOURCES
 *
 * Returns: ETHER
 */
ACTIVITY_EVENT
velodyneStartAcquisitionMain(int* report)
{
  int pthreadCreateResult;

  switch (SDI_F->initParams.mode) {
  case VELODYNE_LIVE:
  case VELODYNE_FROM_FILE:
    if (SDI_F->running != 0) {
      fprintf(stderr, "velodyne: acquistion already started\n");
      *report = S_velodyne_ACQUISITION_ALREADY_STARTED;
      return ETHER;
    }
    if (interface == NULL) {
      fprintf(stderr, "velodyne: interface not open\n");
      *report = S_velodyne_INTERFACE_NOT_OPEN;
      return ETHER;
    }

    /* Initializes some reading parameters. */
    velodyneDestroyBuffer();
    switch (interface->model) {
    case VELODYNE_32:
      velodyneCreateBuffer(VELODYNE_MAX_NUM_PACKETS_PER_REVOLUTION_32);
      break;
    case VELODYNE_64:
      velodyneCreateBuffer(VELODYNE_MAX_NUM_PACKETS_PER_REVOLUTION_64);
      break;
    }
    task.needOneShot = false;
    if (pthread_mutex_init(&task.mtx, NULL)) {
      fprintf(stderr, "velodyne: cannot init mutex\n");
      *report = S_velodyne_NOT_ENOUGH_RESSOURCES;
      return ETHER;
    }
    if (pthread_cond_init(&task.cond, NULL)) {
      fprintf(stderr, "velodyne: cannot init condition variable\n");
      pthread_mutex_destroy(&task.mtx);
      *report = S_velodyne_NOT_ENOUGH_RESSOURCES;
      return ETHER;
    }

    /* Creates the reading thread. SDI_F->running must be set to 1 before
       creating the thread, because if the thread starts while running is still
       0 it will stop. */
    SDI_F->running = 1;
    if (interface->type == VELODYNE_NETWORK) {
      pthreadCreateResult = pthread_create(&task.thr, NULL, velodyneReadLive,
                                           NULL);
    } else {
      pthreadCreateResult = pthread_create(&task.thr,
                                           NULL,
                                           velodyneReadPlayback,
                                           (void*) posFileId);
    }
    if (pthreadCreateResult) {
      fprintf(stderr, "velodyne: cannot create read thread\n");
      pthread_cond_destroy(&task.cond);
      pthread_mutex_destroy(&task.mtx);
      *report = S_velodyne_NOT_ENOUGH_RESSOURCES;
      SDI_F->running = 0;
      return ETHER;
    }
    break;
  case VELODYNE_READ:
    break;
  }

  return ETHER;
}

/*------------------------------------------------------------------------------
 * FOVRemoveHeadingSector - Main
 *
 * Description: selects a heading sector to be removed from the FOV.
 *
 * Reports: BAD_FOV_PARAMS
 *          INTERFACE_NOT_OPEN
 *
 * Returns: ETHER
 */
ACTIVITY_EVENT
velodyneFOVRemoveHeadingSectorMain(
    velodyneFOVRemoveHeadingSectorParams* fovRemoveHeadingSectorParams,
    int* report)
{
  double yawMainToSensorDeg = -SDI_F->sensorToMain.yaw * 180.0 / M_PI;
  int result;
  /* Checks wether interface is open. */
  if (interface == NULL) {
    fprintf(stderr, "velodyne: interface not open\n");
    *report = S_velodyne_INTERFACE_NOT_OPEN;
    return ETHER;
  }
  /* Carries out the request. */
  result = velodyne_add_heading_sector_to_fov_filter(
               interface->model,
               fovRemoveHeadingSectorParams->firstSectorAngle
                   + yawMainToSensorDeg,
               fovRemoveHeadingSectorParams->lastSectorAngle
                   + yawMainToSensorDeg);
  if (result == VELODYNE_ERR_FOV) {
    fprintf(stderr, "velodyne: bad heading sector\n");
    *report = S_velodyne_BAD_FOV_PARAMS;
  }
  return ETHER;
}

/*------------------------------------------------------------------------------
 * FOVRemoveLine - Main
 *
 * Description: selects a line to be removed from the FOV.
 *
 * Reports: BAD_FOV_PARAMS
 *          INTERFACE_NOT_OPEN
 *
 * Returns: ETHER
 */
ACTIVITY_EVENT
velodyneFOVRemoveLineMain(velodyneFOVRemoveLineParams* fovRemoveLineParams,
                          int* report)
{
  int line;
  int result;
  /* Checks wether interface is open. */
  if (interface == NULL) {
    fprintf(stderr, "velodyne: interface not open\n");
    *report = S_velodyne_INTERFACE_NOT_OPEN;
    return ETHER;
  }
  /* Carries out the request. */
  for (line = 0; line < VELODYNE_3D_IMAGE_HEIGHT; line++)
    if (imageParams.lineOrder[line] == fovRemoveLineParams->line)
      break;
  result = velodyne_add_laser_to_fov_filter(interface->model, line);
  if (result == VELODYNE_ERR_FOV) {
    fprintf(stderr, "velodyne: bad line\n");
    *report = S_velodyne_BAD_FOV_PARAMS;
  }
  return ETHER;
}

/*------------------------------------------------------------------------------
 * OneShot - Start
 *
 * Description: checks the state and cleans the target and image structures.
 *
 * Reports: ACQUISITION_NOT_STARTED
 *          INTERFACE_NOT_OPEN
 *
 * Returns: EXEC ETHER
 */
ACTIVITY_EVENT
velodyneOneShotStart(velodyneOneShotParams* oneShotParams, int* report)
{
  switch (SDI_F->initParams.mode) {
  case VELODYNE_LIVE:
  case VELODYNE_FROM_FILE:
    /* Checks the state. */
    if (SDI_F->running == 0) {
      fprintf(stderr, "velodyne: the acquisition is not started\n");
      *report = S_velodyne_ACQUISITION_NOT_STARTED;
      return ETHER;
    }
    if (interface == NULL) {
      fprintf(stderr, "velodyne: the interface is not open\n");
      *report = S_velodyne_INTERFACE_NOT_OPEN;
      return ETHER;
    }

    /* Cleans the target. */
    if (SDI_F->targetDetectionParams.detectionOn == VELODYNE_TRUE)
      memset(&target, 0, sizeof(velodyneTarget));

    break;
  case VELODYNE_READ:
    /* Composes the names of the 3D image and of the position files. */
    sprintf(readImageFileName, "%s.%04d",
            SDI_F->readModeParams.baseName, SDI_F->readModeParams.num);
    sprintf(readPosFileName, "%s.pos.%04d",
            SDI_F->readModeParams.baseName, SDI_F->readModeParams.num);
    break;
  }

  /* Cleans the 3D image. */
  threeDImage->width = 0;
  memset(&threeDImage->position, 0, sizeof(POM_SENSOR_POS));
  memset(threeDImage->points, 0, VELODYNE_3D_IMAGE_HEIGHT
                                 * VELODYNE_3D_IMAGE_WIDTH
                                 * sizeof(velodyne3DPoint));
  return EXEC;
}

/*------------------------------------------------------------------------------
 * OneShot - Main
 *
 * Description: fills the image with one shot of data.
 *
 * Reports: CANNOT_ADD_PACKET
 *          NUM_PACKETS_EXCEEDS_BUFFER_SIZE
 *          CANNOT_OPEN_3D_IMAGE_FILE
 *          CANNOT_READ_3D_IMAGE_FILE
 *          BAD_3D_IMAGE_FILE_FORMAT
 *          BAD_3D_IMAGE_SIZE
 *          CANNOT_READ_POS_FILE
 *
 * Returns: END ETHER
 */
ACTIVITY_EVENT
velodyneOneShotMain(velodyneOneShotParams* oneShotParams, int* report)
{
  velodyneReaderResult grabOneShotResult, readResult;
  POM_SENSOR_POS pos;
  /* struct timeval tv_start, tv_end; */

  switch (SDI_F->initParams.mode) {
  case VELODYNE_LIVE:
  case VELODYNE_FROM_FILE:
    /* Makes sure that velodyneGrabOneShot is accessed in an atomic way. */
    task.needOneShot = true;
    pthread_mutex_lock(&task.mtx);
    grabOneShotResult = velodyneGrabOneShot(&task.buffer,
                                            oneShotParams,
                                            threeDImage,
                                            &imageParams,
                                            &(SDI_F->targetDetectionParams),
                                            &(SDI_F->targetParams),
                                            &target);
    pthread_cond_signal(&task.cond);
    task.needOneShot = false;
    pthread_mutex_unlock(&task.mtx);
    if (grabOneShotResult == VELODYNE_READER_ERR_CANNOT_ADD_PACKET) {
      fprintf(stderr, "velodyne: cannot add packet\n");
      *report = S_velodyne_CANNOT_ADD_PACKET;
      return ETHER;
    } else if (grabOneShotResult
              == VELODYNE_READER_ERR_NUM_PACKETS_EXCEEDS_BUFFER_SIZE) {
      fprintf(stderr, "velodyne: number of packets exceeds buffer size\n");
      *report = S_velodyne_NUM_PACKETS_EXCEEDS_BUFFER_SIZE;
      return ETHER;
    }
    break;
  case VELODYNE_READ:
    /* Reads the 3D image. */
    readResult = velodyne3DImageRead(threeDImage, readImageFileName);
    if (readResult != VELODYNE_READER_OK) {
      switch (readResult) {
      case VELODYNE_READER_ERR_CANNOT_OPEN_3D_IMAGE_FILE:
        fprintf(stderr, "velodyne: cannot open 3D image file\n");
        *report = S_velodyne_CANNOT_OPEN_3D_IMAGE_FILE;
        break;
      case VELODYNE_READER_ERR_CANNOT_READ_3D_IMAGE_FILE:
        fprintf(stderr, "velodyne: cannot read 3D image file\n");
        *report = S_velodyne_CANNOT_READ_3D_IMAGE_FILE;
        break;
      case VELODYNE_READER_ERR_BAD_3D_IMAGE_FILE_FORMAT:
        fprintf(stderr, "velodyne: bad 3D image file format\n");
        *report = S_velodyne_BAD_3D_IMAGE_FILE_FORMAT;
        break;
      case VELODYNE_READER_ERR_BAD_3D_IMAGE_SIZE:
        fprintf(stderr, "velodyne: bad 3D image size\n");
        *report = S_velodyne_BAD_3D_IMAGE_SIZE;
        break;
      default:
        break;
      }
      return ETHER;
    }

    /* Reads and sets the position. */
    if (pomReadSensorPos(readPosFileName, &pos) != 0) {
      fprintf(stderr, "velodyne: cannot read position file\n");
      *report = S_velodyne_CANNOT_READ_POS_FILE;
      return ETHER;
    }
    threeDImage->position = pos;
    break;
  }
  return END;
}

/*------------------------------------------------------------------------------
 * OneShot - End
 *
 * Description: processes target and updates posters.
 *
 * Reports: FAILED_TO_ANALYSE_TARGET
 *          CANNOT_CONVERT_T3D
 *          CANNOT_OPEN_FILE
 *          CANNOT_UPDATE_TARGET_TRACKING_POSTER
 *          CANNOT_UPDATE_TARGET_POSTER
 *          CANNOT_UPDATE_3D_IMAGE_POSTER
 *
 * Returns: ETHER FAIL
 */
ACTIVITY_EVENT
velodyneOneShotEnd(velodyneOneShotParams* oneShotParams, int* report)
{
  T3D sensorToMain;
  T3D mainToOrigin;
  float x, y;
  double* matrix;
  ACTIVITY_EVENT targetAnalysis;
  int i;

  switch (SDI_F->initParams.mode) {
  case VELODYNE_LIVE:
  case VELODYNE_FROM_FILE:
    /* Assesses whether the target has been detected or not. */
    if (SDI_F->targetDetectionParams.detectionOn == VELODYNE_TRUE) {
      /* Analyses target. */
      if (velodyne_analyzeTarget(&SDI_F->targetParams,
                                 &SDI_F->targetDetectionParams,
                                 &target)
          != END) {
        fprintf(stderr, "velodyne: velodyne_analyzeTarget failed\n");
        *report = S_velodyne_FAILED_TO_ANALYSE_TARGET;
        return FAIL;
      }

      /* Fills the target poster. */
      if (target.status != VELODYNE_TARGET_DETECTED) {
        targetAbsolutePosition->numRef++;
        targetAbsolutePosition->nbPts = 0;
        targetRelativePosition->numRef++;
        targetRelativePosition->nbPts = 0;
      } else {
        /* Removes the target points from the 3D image, if required. */
        if (SDI_F->targetDetectionParams.eraseTargetPoints == VELODYNE_TRUE)
          velodyne_eraseTargetPoints(threeDImage, &target);

        /* Updates the target relative position. */
        targetRelativePosition->numRef++;
        targetRelativePosition->nbPts = 1;
        t3dInit(&sensorToMain, T3D_BRYAN, T3D_ALLOW_CONVERSION);
        sensorToMain.euler.euler[0]
            = threeDImage->position.sensorToMain.euler.yaw;
        sensorToMain.euler.euler[1]
            = threeDImage->position.sensorToMain.euler.pitch;
        sensorToMain.euler.euler[2]
            = threeDImage->position.sensorToMain.euler.roll;
        sensorToMain.euler.euler[3]
            = threeDImage->position.sensorToMain.euler.x;
        sensorToMain.euler.euler[4]
            = threeDImage->position.sensorToMain.euler.y;
        sensorToMain.euler.euler[5]
            = threeDImage->position.sensorToMain.euler.z;
        if (!t3dConvertTo(T3D_MATRIX, &sensorToMain)) {
          fprintf(stderr, "velodyne: velodyneOneShotEnd : can't convert t3d\n");
          *report = S_velodyne_CANNOT_CONVERT_T3D;
          return FAIL;
        }
        x = target.cog.coordinates[0];
        y = target.cog.coordinates[1];
        matrix = sensorToMain.matrix.matrix;
        targetRelativePosition->points[0].x
            = x * *(matrix) + y * *(matrix + 1) + *(matrix + 3);
        targetRelativePosition->points[0].y
            = x * *(matrix + 4) + y * *(matrix + 5) + *(matrix + 7);

        /* Updates the absolute position. */
        targetAbsolutePosition->numRef++;
        targetAbsolutePosition->nbPts = 1;
        t3dInit(&mainToOrigin, T3D_BRYAN, T3D_ALLOW_CONVERSION);
        mainToOrigin.euler.euler[0]
            = threeDImage->position.mainToOrigin.euler.yaw;
        mainToOrigin.euler.euler[1]
            = threeDImage->position.mainToOrigin.euler.pitch;
        mainToOrigin.euler.euler[2]
            = threeDImage->position.mainToOrigin.euler.roll;
        mainToOrigin.euler.euler[3]
            = threeDImage->position.mainToOrigin.euler.x;
        mainToOrigin.euler.euler[4]
            = threeDImage->position.mainToOrigin.euler.y;
        mainToOrigin.euler.euler[5]
            = threeDImage->position.mainToOrigin.euler.z;
        if (!t3dConvertTo(T3D_MATRIX, &mainToOrigin)) {
          fprintf(stderr, "velodyne: velodyneOneShotEnd : can't convert t3d\n");
          *report = S_velodyne_CANNOT_CONVERT_T3D;
          return FAIL;
        }
        x = targetRelativePosition->points[0].x;
        y = targetRelativePosition->points[0].y;
        matrix = mainToOrigin.matrix.matrix;
        targetAbsolutePosition->points[0].x
            = x * *(matrix) + y * *(matrix + 1) + *(matrix + 3);
        targetAbsolutePosition->points[0].y
            = x * *(matrix + 4) + y * *(matrix + 5) + *(matrix + 7);

        /* Tracks target. */
        if (SDI_F->targetDetectionParams.trackingOn == VELODYNE_TRUE) {
          velodyneTrackTarget(&target,
                              targetTracking,
                              &(SDI_F->targetTrackingParams),
                              &(SDI_F->targetTrackingFileOutputParams),
                              targetAbsolutePosition->points[0].x,
                              targetAbsolutePosition->points[0].y,
                              report);
          if (*report == S_velodyne_CANNOT_OPEN_FILE)
            fprintf(stderr, "velodyneOneShotEnd: cannot open file\n");
          for (i = targetTrackingPoster->nbPts;
               i < targetTracking->numTargets;
               i++) {
            targetTrackingPoster->points[i].x
                = targetTracking->targets[i].cog.coordinates[0];
            targetTrackingPoster->points[i].y
                = targetTracking->targets[i].cog.coordinates[1];
          }
          targetTrackingPoster->nbPts = targetTracking->numTargets;
          if (velodyneTargetTrackingGENPOS_TRAJ_POINTSPosterWrite(
                  VELODYNE_TARGETTRACKING_POSTER_ID, targetTrackingPoster)
              != OK) {
            *report = S_velodyne_CANNOT_UPDATE_TARGET_TRACKING_POSTER;
            return FAIL;
          }
        }
      }

      /* Now writes the target posters. */
      if (velodyneTargetAbsolutePositionGENPOS_TRAJ_POINTSPosterWrite(
              VELODYNE_TARGETABSOLUTEPOSITION_POSTER_ID, targetAbsolutePosition)
          != OK) {
        *report = S_velodyne_CANNOT_UPDATE_TARGET_POSTER;
        return FAIL;
      }
      if (velodyneTargetRelativePositionGENPOS_TRAJ_POINTSPosterWrite(
              VELODYNE_TARGETRELATIVEPOSITION_POSTER_ID, targetRelativePosition)
          != OK) {
        *report = S_velodyne_CANNOT_UPDATE_TARGET_POSTER;
        return FAIL;
      }
    }
    break;
  case VELODYNE_READ:
    /* Resets the 3D image and the position files' names. */
    strcpy(readImageFileName, "");
    strcpy(readPosFileName, "");

    /* Increases the file extension number. */
    SDI_F->readModeParams.num += 1;
    break;
  }

  /* Writes the 3D image poster. */
  if (velodyneThreeDImagevelodyne3DImagePosterWrite(
          VELODYNE_THREEDIMAGE_POSTER_ID, threeDImage)
      != OK) {
    fprintf(stderr, "velodyne: cannot update 3D image poster\n");
    *report = S_velodyne_CANNOT_UPDATE_3D_IMAGE_POSTER;
    return FAIL;
  }
  return ETHER;
}

/*------------------------------------------------------------------------
 * SetTargetDetectionParams
 *
 * Description: allocates the memory required to track the target
 *
 * Reports:      OK
 *              S_velodyne_CAN_NOT_ALLOCATE_MEMORY
 */

/* Returns:  ETHER */

ACTIVITY_EVENT
velodyneSetTargetDetectionParams(velodyneTargetDetectionParams* detectionParams,
				int* report)
{
  /* Just sets the target structure to "0" -- whatever happens */
  memset (&target, 0, sizeof(velodyneTarget));
  return ETHER;
}

/*------------------------------------------------------------------------
 * DumpTarget
 *
 * Description: saves the target points in an ASCII file
 *
 * Reports:      OK
 *              S_velodyne_CANT_OPEN_FILE
 */

/* velodyneDumpTargetMain  -  codel EXEC of DumpTarget
   Returns:  EXEC END ETHER FAIL ZOMBIE */

ACTIVITY_EVENT
velodyneDumpTargetMain(velodyneSaveParams* saveParams,
		       int* report)
{
  char	fileName[VELODYNE_MAX_STRLEN];
  FILE*	file;
  int	i;
  /* opens the file */
  sprintf (fileName, "%s.%04d",
	   saveParams->baseName, saveParams->num);
  file = fopen (fileName, "w");
  if (file == NULL) {
    *report = S_velodyne_CANT_OPEN_FILE;
    return FAIL;
  }
  for (i = 0; i < target.detectedTargetPoints; i++) {
    if ((target.targetPoints + i)->status == VELODYNE_GOOD_3DPOINT)
      fprintf (file, "%.2f %.2f %.2f\n", 
	       (target.targetPoints + i)->coordinates[0],
	       (target.targetPoints + i)->coordinates[1],
	       (target.targetPoints + i)->coordinates[2]);
  }
  fclose (file);
  return ETHER;
}

/*------------------------------------------------------------------------
 * Save
 *
 * Description:
 *
 * Reports: OK
 *          S_velodyne_CANNOT_OPEN_FILE
 *          S_velodyne_CANNOT_WRITE_ON_FILE
 */

/* velodyneSaveMain  -  codel MAIN of Save
   Returns: ETHER */
ACTIVITY_EVENT
velodyneSaveMain(int* report)
{
  int i, j;
  char fileName[VELODYNE_MAX_STRLEN];
  FILE* file;
  velodyne3DPoint* vel3DPoint;
  VELODYNE_CARTESIAN_POINT cartPt;

  /* Opens file */
  sprintf (fileName, "%s.%04d",
           SDI_F->saveParams.baseName, SDI_F->saveParams.num);
  file = fopen (fileName, "w");
  if (file == NULL) {
    *report = S_velodyne_CANNOT_OPEN_FILE;
    return ETHER;
  }

  /* Writes the header */
  fprintf (file, VELODYNE_FILE_FORMAT);
  fprintf (file, "\n%d %d %d\n",
           threeDImage->height, threeDImage->width, threeDImage->maxScanWidth);

  /* Writes the points */
  for (i = 0; i < threeDImage->height; i++) {
    for (j = 0; j < threeDImage->width; j++) {
      vel3DPoint = velodyne3DImageAt (threeDImage, i, j);
      cartPt.x = vel3DPoint->coordinates[0];
      cartPt.y = vel3DPoint->coordinates[1];
      cartPt.z = vel3DPoint->coordinates[2];
      cartPt.intensity = (u_int8_t) vel3DPoint->intensity;
      cartPt.status = (VELODYNE_POINT_STATUS) vel3DPoint->status;
      if (fwrite (&cartPt, sizeof (VELODYNE_CARTESIAN_POINT), 1, file) != 1) {
        *report = S_velodyne_CANNOT_WRITE_ON_FILE;
        fclose (file);
        return ETHER;
      }
    }
  }

  /* Closes file */
  fclose (file);

  /* Composes the file name and saves the pom position */
  sprintf (fileName, "%s.pos.%04d",
           SDI_F->saveParams.baseName, SDI_F->saveParams.num);
  pomWriteSensorPos (fileName, &(threeDImage->position));

  SDI_F->saveParams.num += 1;

  return ETHER;
}

/*------------------------------------------------------------------------
 * SaveIm3D
 *
 * Description:
 *
 * Reports: OK
 *          S_velodyne_CANNOT_OPEN_FILE
 *          S_velodyne_CANNOT_ALLOCATE_IMAGE_3D
 */

/* velodyneSaveIm3DMain  -  codel MAIN of SaveIm3D
   Returns: ETHER */
ACTIVITY_EVENT
velodyneSaveIm3DMain(int* report)
{
  int i, j;
  char fileName[VELODYNE_MAX_STRLEN];
  DATA_IM3D im3D;
  DATA_PT3D* pt3D;
  velodyne3DPoint* vel3DPoint;

  /* Allocates the DATA_IM3D */
  if (init_data_im3d (&im3D, threeDImage->height, threeDImage->maxScanWidth,
                      0, 0, 0)
      != ERR_IM3D_OK) {
    *report = S_velodyne_CANNOT_ALLOCATE_IMAGE_3D;
    return ETHER;
  }

  /* Copies the 3D image into the DATA_IM3D */
  for (i = 0; i < threeDImage->height; i++) {
    for (j = 0; j < threeDImage->width; j++) {
      pt3D = data_im3d_pt (&im3D, i, j);
      vel3DPoint = velodyne3DImageAt (threeDImage, i, j);
      if (vel3DPoint->status == VELODYNE_GOOD_3DPOINT) {
        pt3D->state = GOOD_PT3D;
        pt3D->coord_1 = vel3DPoint->coordinates[0];
        pt3D->coord_2 = vel3DPoint->coordinates[1];
        pt3D->coord_3 = vel3DPoint->coordinates[2];
      }
    }
  }

  im3D.header.coordonnees = COORDONNEES_CARTESIENNES;
  im3D.header.capteur = CAPTEUR_LASER1;
  im3D.header.repere = REPERE_CAPTEUR;
  im3D.header.points = POINTS_CALCULES;

  /* Composes the file name and saves the DATA_IM3D */
  sprintf (fileName, "%s.%04d",
           SDI_F->saveIm3DParams.baseName, SDI_F->saveIm3DParams.num);
  if (write_data_im3d_binaire (&im3D, fileName) != ERR_IM3D_OK) {
    *report = S_velodyne_CANNOT_OPEN_FILE;
    return ETHER;
  }

  /* Composes the file name and saves the pom position */
  sprintf (fileName, "%s.pos.%04d",
           SDI_F->saveIm3DParams.baseName, SDI_F->saveIm3DParams.num);
  pomWriteSensorPos (fileName, &(threeDImage->position));

  SDI_F->saveIm3DParams.num += 1;

  /* Frees the im3d */
  empty_data_im3d (&im3D);

  return ETHER;
}
